/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl.abstr;

import dk.sds.ecpr2.dao.CrossTableDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.*;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.domain.abstr.PersonDescriptorWithServiceActorVO;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.CollectionUtils;

import javax.annotation.PostConstruct;
import javax.sql.DataSource;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

@Repository
public abstract class AbstractCrossTableDAOImpl implements CrossTableDAO {
    protected NamedParameterJdbcTemplate namedParamJdbcTemplate;
    @Autowired
    private DataSource dataSource;

    @PostConstruct
    private void init() {
        namedParamJdbcTemplate = new NamedParameterJdbcTemplate(dataSource);
    }


    private static final Logger log = LogManager.getLogger(AbstractCrossTableDAOImpl.class);

    @Autowired
    PersonDAOImpl personDAO;

    @Autowired
    private IdentifierDAOImpl identifierDAO;

    @Autowired
    private AddressDAOImpl addressDAO;

    @Autowired
    private BirthDateDAOImpl birthDateDAO;

    @Autowired
    private ContactDAOImpl contactDAO;

    @Autowired
    private GenderDAOImpl genderDAO;

    @Autowired
    private NameDAOImpl nameDAO;

    @Override
    public List<PersonVO> getPersonsWithRelations(List<String> uuids) throws DAOException {
        List<PersonVO> persons = personDAO.getPersonsByUUIDs(uuids);
        addRelatedInformationToPersons(persons);
        return persons;
    }

    @Override
    public List<PersonVO> getPersonWithRelations(String uuid) throws DAOException {
        List<PersonVO> person = personDAO.getPersonsByUUIDs(Collections.singletonList(uuid));
        addRelatedInformationToPersons(person);
        return person;
    }
  
    @Override
    public List<PersonVO> getRecentlyReservedPersons(Instant date) throws DAOException {
        List<PersonVO> persons = personDAO.getReservedPersonByModifiedTime(date);
        addRelatedInformationToPersons(persons);
        return persons;
    }

    @Override
    public List<PersonVO> getPersonsWithRelationsByPersonId(List<PersonIdVO> identifiers) throws DAOException {
        List<PersonVO> persons = personDAO.getPersonUuidByPersonIds(identifiers);
        if (CollectionUtils.isEmpty(persons)) {
            return Collections.emptyList();
        }
        addRelatedInformationToPersonsByPersonPid(persons);
        return persons;
    }

    @Transactional
    @Override
    public void deleteReservedPersonIds(Instant date) throws DAOException {
        List<PersonVO> personList = personDAO.getPersonByReservedUntil(date);
        List<Integer> personPIDs = new ArrayList<>();
        if (!personList.isEmpty()) {
            for (PersonVO persons : personList) {
                personPIDs.add(persons.getPersonPID());
            }
            log.debug("Done fetching data for exceeded reserved Person Ids (" + personList.size() + " rows). Deleting...");
            try {
                final String tableName = identifierDAO.getTableName();
                String sql = "DELETE FROM " + tableName + " WHERE personPID" + " IN (:Ids)";
                final MapSqlParameterSource parameterMap = new MapSqlParameterSource();
                parameterMap.addValue("Ids", personPIDs);

                namedParamJdbcTemplate.update(sql, parameterMap);

            } catch (Exception e) {
                throw new DAOException("An error occurred updating valid-to timestamps", e);
            }
            try {
                final String tableName = genderDAO.getTableName();
                String sql = "DELETE FROM " + tableName + " WHERE personPID" + " IN (:Ids)";
                final MapSqlParameterSource parameterMap = new MapSqlParameterSource();
                parameterMap.addValue("Ids", personPIDs);

                namedParamJdbcTemplate.update(sql, parameterMap);

            } catch (Exception e) {
                throw new DAOException("An error occurred updating valid-to timestamps", e);
            }
            try {
                final String tableName = birthDateDAO.getTableName();
                String sql = "DELETE FROM " + tableName + " WHERE personPID" + " IN (:Ids)";
                final MapSqlParameterSource parameterMap = new MapSqlParameterSource();
                parameterMap.addValue("Ids", personPIDs);

                namedParamJdbcTemplate.update(sql, parameterMap);

            } catch (Exception e) {
                throw new DAOException("An error occurred updating valid-to timestamps", e);
            }
            try {
                final String tableName = personDAO.getTableName();
                String sql = "DELETE FROM " + tableName + " WHERE personPID" + " IN (:Ids)";
                final MapSqlParameterSource parameterMap = new MapSqlParameterSource();
                parameterMap.addValue("Ids", personPIDs);

                namedParamJdbcTemplate.update(sql, parameterMap);

            } catch (Exception e) {
                throw new DAOException("An error occurred updating valid-to timestamps", e);
            }
            log.debug("Done deleting all data for exceeded reserved Person Ids (" + personList.size() + " rows).");
        }
    }

    @Override
    public abstract List<PersonVO> getPersonsWithRelationsAfterPID(int minPersonPid,
                                                          int batchSize,
                                                          boolean excludeReservedUntilFilter,
                                                          boolean excludeDuplicatePersonsFilter) throws DAOException;

    private void addRelatedInformationToPersonsByPersonPid(List<PersonVO> persons) throws DAOException {
        Map<Integer, PersonVO> personMap = getPersonMap(persons);

        List<Integer> personPids = new ArrayList<>();
        for (PersonVO person : persons) {
            personPids.add(person.getPersonPID());
        }
        addInformationToPerson(personMap, Collections.emptyList(), personPids);
    }

    private Map<Integer, PersonVO> getPersonMap(List<PersonVO> persons) {
        Map<Integer, PersonVO> personMap = new HashMap<>();
        for (PersonVO p : persons) personMap.put(p.getPersonPID(), p);
        return personMap;
    }

    private void addRelatedInformationToPersons(List<PersonVO> persons) throws DAOException {
        Map<Integer, PersonVO> personMap = new HashMap<>();
        for (PersonVO p : persons) personMap.put(p.getPersonPID(), p);

        List<String> personUuids = new ArrayList<>();
        for (PersonVO person : persons) {
            personUuids.add(person.getPersonUUID());
        }
        addInformationToPerson(personMap, personUuids, Collections.emptyList());
    }

    private void addInformationToPerson(Map<Integer, PersonVO> personMap, List<String> personUuids, List<Integer> personPids) throws DAOException {
        List<IdentifierVO> identifiers = identifierDAO.getWithServiceActor(personUuids, personPids);
        for (IdentifierVO o : identifiers) {
            personMap.get(o.getPersonPID()).getIdentifiers().add(o);
        }
        for (NameVO o : nameDAO.getWithServiceActor(personUuids, personPids)) {
            personMap.get(o.getPersonPID()).getNames().add(o);
        }
        for (GenderVO o : genderDAO.getWithServiceActor(personUuids, personPids)) {
            personMap.get(o.getPersonPID()).getGenders().add(o);
        }
        for (BirthDateVO o : birthDateDAO.getWithServiceActor(personUuids, personPids)) {
            personMap.get(o.getPersonPID()).getBirthDates().add(o);
        }
        for (ContactVO o : contactDAO.getWithServiceActor(personUuids, personPids)) {
            personMap.get(o.getPersonPID()).getContacts().add(o);
        }
        for (AddressVO o : addressDAO.getWithServiceActor(personUuids, personPids)) {
            personMap.get(o.getPersonPID()).getAddresses().add(o);
        }
    }

    protected  <T extends PersonDescriptorWithServiceActorVO> List<T> getNewestPersonDescriptors(List<T> unfiltered, boolean propertyExcludeFilter) {
        if (!propertyExcludeFilter) {
            return unfiltered;
        }
        return unfiltered.stream().filter(p1 -> {
            PersonDescriptorWithServiceActorVO p2Max = unfiltered.stream()
                    .filter(p2 -> !p2.getPersonPID().equals(p1.getPersonPID()))
                    .filter(p2 -> p2.getUuid().equals(p1.getUuid()))
                    .max(Comparator.comparing(PersonDescriptorWithServiceActorVO::getValidFrom))
                    .orElse(null);
            if (p2Max == null) {
                return true;
            }
            return p1.getValidFrom().after(p2Max.getValidFrom());
        }).collect(Collectors.toList());
    }

    protected List<PersonVO> getNewestPersons(List<PersonVO> persons) {
        return persons.stream().filter(p1 -> {
            PersonVO p2Max = persons.stream()
                    .filter(p2 -> !p2.getPersonPID().equals(p1.getPersonPID()))
                    .filter(p2 -> p2.getPersonUUID().equals(p1.getPersonUUID()))
                    .max(Comparator.comparing(PersonVO::getVersion))
                    .orElse(null);
            if (p2Max == null) {
                return true;
            }
            return p1.getVersion() > p2Max.getVersion();
        }).collect(Collectors.toList());
    }

}
