/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.datagenerator;

import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.datagenerator.db.local.LocalContext;
import dk.sds.ecpr2.datagenerator.db.local.LocalContextFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.PropertySource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.sql.DataSource;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Component
@ComponentScan(basePackages = {"dk.sds.ecpr2.dao", "dk.sds.ecpr2.util", "dk.sds.ecpr2.datagenerator"})
@PropertySource({"classpath:ecpr2.properties"})
@PropertySource({"classpath:${datageneration.properties.file}"})
public class DataGeneratorApplication {

    public static final double ACCEPTABLE_NUMBER_OF_ROWS_DIFFERENCE_IN_PERCENT = 5.0;

    @Autowired
    private DataSource dataSource;

    @Autowired
    private List<AbstractGenerator> generatorsInReferentialOrder;

    private String filename = "identifiers.csv";

    private Map<String, AbstractGenerator.IdRange> generatedIdRanges = new HashMap<>();

    private double startTime;

    public static void main(String... args) throws DAOException, IOException {
        new DataGeneratorApplication().run();
    }

    public void run() throws DAOException, IOException {
        LocalContext ctx = LocalContextFactory.createLocalContext("com.mysql.cj.jdbc.Driver");
        ctx.addDataSource("java:jboss/datasources/ecpr2","jdbc:mysql://db:3306/ecpr2", "ecpr2_service", "Password1");

        ApplicationContext applicationContext = new AnnotationConfigApplicationContext(DataGeneratorApplication.class);
        applicationContext.getAutowireCapableBeanFactory().autowireBean(this);

        System.out.println("Running data generator application.");

        if (!allTablesAlreadyHaveAcceptableRowCount()) {
            removeAllData();
            generateData();
        }

        NamedParameterJdbcTemplate namedParamJdbcTemplate = new NamedParameterJdbcTemplate(dataSource);
        List<String> identifiers = namedParamJdbcTemplate.queryForList("SELECT Value FROM ecpr2.Identifier WHERE ValidTo > now() AND OID = '1.2.208.176.1.6.1.1'", new HashMap<>(), String.class);
//        System.out.println(identifiers);

        BufferedWriter outputWriter = new BufferedWriter(new FileWriter(filename));
        for (String identifier : identifiers) {
            outputWriter.write(identifier);
            outputWriter.newLine();
        }
        outputWriter.flush();
        outputWriter.close();
    }

    public void generateData() throws DAOException {
        printAndTime("Starting data generation.");

        for (AbstractGenerator generator : generatorsInReferentialOrder) {
            AbstractGenerator.IdRange generatedIds = generator.generate(generatedIdRanges);
            generatedIdRanges.put(generator.getTableName(), generatedIds);
            printAndTime(generator.getNumberOfRowsToGenerate() + " " + generator.getTableName() + " rows generated.");
        }
        printAndTime("Successfully completed data generation.");
    }

    public void removeAllData() throws DAOException {
        printAndTime("Deleting data.");
        for (int i = generatorsInReferentialOrder.size() - 1; i >= 0; i--) {
            AbstractGenerator generator = generatorsInReferentialOrder.get(i);
            generator.removeRows();
            printAndTime("Removed data from " + generator.getTableName() + " table starting from id " + generator.getMinimumIdToRemove());
        }
        printAndTime("Successfully completed deletion.");
    }

    private boolean allTablesAlreadyHaveAcceptableRowCount() throws DAOException {
        boolean amountOfDataIsOk = true;
        printAndTime("Checking current amount of data.");

        for (AbstractGenerator generator : generatorsInReferentialOrder) {
            int actualRows = generator.rowCount();
            int neededRows = generator.getNumberOfRowsToGenerate();
            double percentOff = 100.0 * Math.abs(actualRows - neededRows) / neededRows;
            if (percentOff > ACCEPTABLE_NUMBER_OF_ROWS_DIFFERENCE_IN_PERCENT) {
                printAndTime("Wrong amount of data in the " + generator.getTableName() + "table. Found: " + actualRows + " Expected: " + neededRows + " (" + percentOff + "% off)");
                amountOfDataIsOk = false;
                break;
            } else {
                printAndTime("Table " + generator.getTableName() + " had " + actualRows + ". Expected " + neededRows + " (" + percentOff + "% off) OK");
            }

        }

        printAndTime("Done checking. Amount of data was " + (amountOfDataIsOk ? "ok." : "not ok."));
        return amountOfDataIsOk;
    }

    protected void printAndTime(String message) {
        if (startTime > 0) {
            final double durationMillis = (System.nanoTime() - startTime) / 1000000;
            String durationText = durationMillis > 1000
                    ? Math.round(durationMillis / 1000) + " s"
                    : Math.round(durationMillis) + " ms";
            System.out.println(message + "  (Time since last log: " + durationText + ")");
        } else {
            System.out.println(message);
        }
        startTime = System.nanoTime();
    }

}
