/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl.abstr;

import dk.sds.ecpr2.dao.ECPR2DAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.PIDRowMapper;
import dk.sds.ecpr2.domain.abstr.PersonDescriptorVO;
import dk.sds.ecpr2.util.TimeHelper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;

import java.time.Instant;
import java.util.*;


/**
 * This class contains general purpose database operations that only make sense within the E-CPR project.
 */
public abstract class AbstractECPRDAOImpl<T extends PersonDescriptorVO> extends AbstractDAOImpl<T> implements ECPR2DAO {

    public static final Date VALID_TO_DEFAULT = TimeHelper.createNew(2999, Calendar.DECEMBER, 31);
    public static final String ACTOR_FIELDS = "a.ServiceActorPID, a.PersonIdentifier as ActorPersonIdentifier, a.PersonIdentifierOID as ActorPersonIdentifierOID, a.FirstName, a.LastName, a.Role, a.AuthorisationNumber, a.OrganizationId, a.OrganizationName, a.HashValue, a.ModifiedTime";
    protected static final String TIMESTAMP_AND_ACTOR_FIELDS = "p.ModifiedTime, " + ACTOR_FIELDS;

    protected List<T> getWithServiceActorByPersonUUIDs(List<String> personUuids, List<Integer> personPids) throws DAOException {
        if (personUuids.isEmpty() && personPids.isEmpty()) {
            return new ArrayList<T>();
        }
        try {
            String columnName;
            final boolean uuidIsInitialized = (!personUuids.isEmpty());
            List columnList;
            if (uuidIsInitialized) {
                columnName = "PersonUUID";
                columnList = personUuids;
            } else {
                columnName = "PersonPid";
                columnList = personPids;
            }
            final String tableName = getTableName();
            String sql = "SELECT t.*, p.PersonUUID, " + TIMESTAMP_AND_ACTOR_FIELDS + " FROM ecpr2.Person p, ecpr2.ServiceActor a, "
                    + tableName + " t WHERE p." + columnName + " IN (:list)" +
                    " AND t.PersonPID = p.PersonPID AND p.ServiceActorPID = a.ServiceActorPID";

            final MapSqlParameterSource parms = new MapSqlParameterSource()
                    .addValue("list", columnList);
            return namedParamJdbcTemplate.query(sql, parms, getDefaultRowMapper());
        } catch (Exception e) {
            throw new DAOException("An error occurred loading " + getTableName() + " by UUIDs", e);
        }
    }

    protected List<T> getAllFieldsBetweenPersonPIDWithServiceActor(int minPersonPID, int maxPersonPID) throws DAOException {
        if (minPersonPID < 0 || maxPersonPID < 0) {
            throw new UnsupportedOperationException("Min and Max Person PID must be specified.");
        }
        try {
            final String tableName = getTableName();
            String sql = "SELECT t.*, p.PersonUUID, " + ACTOR_FIELDS + " FROM ecpr2.Person p, ecpr2.ServiceActor a, " + tableName + " t ";
            final MapSqlParameterSource parms = new MapSqlParameterSource();
            sql += " WHERE t.PersonPID >= :minPersonPid AND t.PersonPID <= :maxPersonPid " +
                    "AND t.PersonPID = p.PersonPID AND p.ServiceActorPID = a.ServiceActorPID";
            parms.addValue("minPersonPid", minPersonPID);
            parms.addValue("maxPersonPid", maxPersonPID);

            final PersonDescriptorRowMapper<T> rowMapper = (PersonDescriptorRowMapper<T>) getDefaultRowMapper();
            rowMapper.useOnlySearchFields = false;
            return namedParamJdbcTemplate.query(sql, parms, rowMapper);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading selected fields from " + getTableName(), e);
        }
    }

    protected List<T> getAllFieldsAfterModifiedTimeWithServiceActor(Instant modifiedTime) throws DAOException {
        if (modifiedTime == null) {
            throw new UnsupportedOperationException("ModifiedTime must not be null");
        }
        try {
            final String tableName = getTableName();
            String sql = "SELECT t.*, p.PersonUUID, " + ACTOR_FIELDS + " FROM ecpr2.Person p, ecpr2.ServiceActor a, " + tableName + " t ";
            final MapSqlParameterSource parms = new MapSqlParameterSource();
            sql += " WHERE t.ModifiedTime >= :modifiedTime " +
                    "AND t.PersonPID = p.PersonPID AND p.ServiceActorPID = a.ServiceActorPID";
            parms.addValue("modifiedTime", modifiedTime);

            final PersonDescriptorRowMapper<T> rowMapper = (PersonDescriptorRowMapper<T>) getDefaultRowMapper();
            rowMapper.useOnlySearchFields = false;
            return namedParamJdbcTemplate.query(sql, parms, rowMapper);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading selected fields from " + getTableName(), e);
        }
    }

    protected List<T> getWithServiceActorByPersonUUIDAndDate(String personUuid, Date date) throws DAOException {
        try {
            final String tableName = getTableName();
            String sql = "SELECT t.*, p.PersonUUID," + TIMESTAMP_AND_ACTOR_FIELDS + " FROM ecpr2.Person p, ecpr2.ServiceActor a, "
                    + tableName + " t WHERE p.PersonUUID = :personUuid" +
                    " AND t.PersonPID = p.PersonPID AND t.ValidFrom <= :date AND t.ValidTo > :date AND p.ServiceActorPID = a.ServiceActorPID";

            final MapSqlParameterSource parameterMap = new MapSqlParameterSource()
                    .addValue("personUuid", personUuid)
                    .addValue("date", date);
            return namedParamJdbcTemplate.query(sql, parameterMap, getDefaultRowMapper());
        } catch (Exception e) {
            throw new DAOException("An error occurred loading " + getTableName() + " by UUID and date", e);
        }
    }

    protected void assignFieldsForNew(PersonDescriptorVO vo, Date nowTimestamp, int personId) {
        if (vo.getUuid() == null) {
            vo.setUuid(UUID.randomUUID().toString());
        }
        if (personId != 0) {
            vo.setPersonPID(personId);
        }
        if (nowTimestamp != null) {
            vo.setValidFrom(nowTimestamp);
            vo.setModified(nowTimestamp);
        }
        if (vo.getValidTo() == null) {
            vo.setValidTo(AbstractECPRDAOImpl.VALID_TO_DEFAULT);
        }
    }

    @Override
    public void delete(PersonDescriptorVO vo, Date timestamp) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), timestamp);
    }

    public void updateValidToTimestamp(String uuid, Date timestamp) throws DAOException {
        final String sql = "SELECT " + getIdColumnName() + " FROM " + getTableName() + " WHERE Uuid = :uuid ORDER BY ValidFrom DESC LIMIT 1";
        final MapSqlParameterSource parameters = new MapSqlParameterSource().addValue("uuid", uuid);
        List<Integer> result = namedParamJdbcTemplate.query(sql, parameters, new PIDRowMapper(getIdColumnName()));
        if (result.size() != 1) {
            throw new DAOException("Expected single result for latest record from " + getTableName() + " and uuid " + uuid);
        }
        Integer pidToUpdate =  result.get(0);
        if (pidToUpdate == null) {
            return;
        }

        try {
            final String tableName = getTableName();

            String updateSql = "UPDATE " + tableName + " SET ValidTo = :ValidTo, ModifiedTime = :ModifiedTime";
            updateSql += " WHERE " + getIdColumnName() + " = :Id";
            final MapSqlParameterSource updateParameters = new MapSqlParameterSource();
            updateParameters.addValue("ValidTo", timestamp);
            updateParameters.addValue("ModifiedTime", timestamp);
            updateParameters.addValue("Id", pidToUpdate);
            namedParamJdbcTemplate.update(updateSql, updateParameters);
        } catch (Exception e) {
            throw new DAOException("An error occurred updating valid-to timestamps", e);
        }
    }
}
