/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.domain;


import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonInclude;
import dk.sds.ecpr2.Modified;
import dk.sds.ecpr2.constants.SearchConstants;
import dk.sds.ecpr2.domain.abstr.PersonDescriptorWithServiceActorVO;

import java.util.*;

@JsonInclude(JsonInclude.Include.NON_EMPTY)
public class PersonVO extends PersonDescriptorWithServiceActorVO {

    private KeyIdentifierVO keyIdentifier;
    @JsonIgnore
    private Integer personPID;
    @JsonIgnore
    private Integer serviceActorPID;
    @JsonIgnore
    private Integer modifierPID;
    private String personUUID;
    private Integer version;
    private String versionDescription;
    private Date reservedUntil;
    private Date modifiedTime;
    private String status;

    private List<Modified> modifiers = new ArrayList<>();
    private List<IdentifierVO> identifiers = new ArrayList<>();
    private List<NameVO> names = new ArrayList<>();
    private List<GenderVO> genders = new ArrayList<>();
    private List<BirthDateVO> birthDates = new ArrayList<>();
    private List<ContactVO> contacts = new ArrayList<>();
    private List<AddressVO> addresses = new ArrayList<>();

    @JsonIgnore
    public List<LuceneKeyValue> getLuceneList() {
        List<LuceneKeyValue> luceneKeyValueList = new ArrayList<>();
        luceneKeyValueList.add(new LuceneKeyValue(SearchConstants.PERSON_UUID, personUUID));
        for (NameVO name : names) {
            luceneKeyValueList.addAll(name.getLuceneList());
        }
        for (GenderVO gender : genders) {
            luceneKeyValueList.addAll(gender.getLuceneList());
        }
        for (BirthDateVO birthDate : birthDates) {
            luceneKeyValueList.addAll(birthDate.getLuceneList());
        }
        for (ContactVO contact : contacts) {
            luceneKeyValueList.addAll(contact.getLuceneList());
        }
        for (AddressVO address : addresses) {
            luceneKeyValueList.addAll(address.getLuceneList());
        }

        return luceneKeyValueList;
    }

    public static int getMaxPersonPID(List<PersonVO> personBatch) {
        try {
            return personBatch
                    .stream()
                    .mapToInt(PersonVO::getPersonPID)
                    .max()
                    .orElse(0);
        } catch (NullPointerException ignored) { // max() returns NPE when max is null
            return 0;
        }
    }

    public static int getMinPersonPID(List<PersonVO> personBatch) {
        try {
            return personBatch
                    .stream()
                    .mapToInt(PersonVO::getPersonPID)
                    .min()
                    .orElse(Integer.MAX_VALUE);
        } catch (NullPointerException ignored) { // min() returns NPE when min is null
            return Integer.MAX_VALUE;
        }
    }

    public KeyIdentifierVO getKeyIdentifier() {
        return keyIdentifier;
    }

    public void setKeyIdentifier(KeyIdentifierVO keyIdentifier) {
        this.keyIdentifier = keyIdentifier;
    }

    public Integer getPersonPID() {
        return personPID;
    }

    public void setPersonPID(Integer personPID) {
        this.personPID = personPID;
    }

    public Integer getServiceActorPID() {
        return serviceActorPID;
    }

    public void setServiceActorPID(Integer serviceActorPID) {
        this.serviceActorPID = serviceActorPID;
    }

    public Integer getModifierPID() {
        return modifierPID;
    }

    public void setModifierPID(Integer modifierPID) {
        this.modifierPID = modifierPID;
    }

    public String getPersonUUID() {
        return personUUID;
    }

    public void setPersonUUID(String personUUID) {
        this.personUUID = personUUID;
    }

    public Integer getVersion() {
        return version;
    }

    public void setVersion(Integer version) {
        this.version = version;
    }

    public String getVersionDescription() {
        return versionDescription;
    }

    public void setVersionDescription(String versionDescription) {
        this.versionDescription = versionDescription;
    }

    public Date getReservedUntil() {
        return reservedUntil;
    }

    public void setReservedUntil(Date reservedUntil) {
        this.reservedUntil = reservedUntil;
    }

    public Date getModifiedTime() {
        return modifiedTime;
    }

    public void setModifiedTime(Date modifiedTime) {
        this.modifiedTime = modifiedTime;
    }

    public String getStatus() {
        return status;
    }

    public void setStatus(String status) {
        this.status = status;
    }

    public List<Modified> getModifiers() {
        return modifiers;
    }

    public void setModifiers(List<Modified> modifiers) {
        this.modifiers = modifiers;
    }

    public List<IdentifierVO> getIdentifiers() {
        return identifiers;
    }

    public void setIdentifiers(List<IdentifierVO> identifiers) {
        this.identifiers = identifiers;
    }

    public List<NameVO> getNames() {
        return names;
    }

    public void setNames(List<NameVO> names) {
        this.names = names;
    }

    public List<GenderVO> getGenders() {
        return genders;
    }

    public void setGenders(List<GenderVO> genders) {
        this.genders = genders;
    }

    public List<BirthDateVO> getBirthDates() {
        return birthDates;
    }

    public void setBirthDates(List<BirthDateVO> birthDates) {
        this.birthDates = birthDates;
    }

    public List<ContactVO> getContacts() {
        return contacts;
    }

    public void setContacts(List<ContactVO> contacts) {
        this.contacts = contacts;
    }

    public List<AddressVO> getAddresses() {
        return addresses;
    }

    public void setAddresses(List<AddressVO> addresses) {
        this.addresses = addresses;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        if (!super.equals(o)) return false;
        PersonVO personVO = (PersonVO) o;
        return Objects.equals(keyIdentifier, personVO.keyIdentifier) && Objects.equals(personPID, personVO.personPID) && Objects.equals(serviceActorPID, personVO.serviceActorPID) && Objects.equals(modifierPID, personVO.modifierPID) && Objects.equals(personUUID, personVO.personUUID) && Objects.equals(version, personVO.version) && Objects.equals(versionDescription, personVO.versionDescription) && Objects.equals(reservedUntil, personVO.reservedUntil) && Objects.equals(modifiedTime, personVO.modifiedTime) && Objects.equals(status, personVO.status) && Objects.equals(modifiers, personVO.modifiers) && Objects.equals(identifiers, personVO.identifiers) && Objects.equals(names, personVO.names) && Objects.equals(genders, personVO.genders) && Objects.equals(birthDates, personVO.birthDates) && Objects.equals(contacts, personVO.contacts) && Objects.equals(addresses, personVO.addresses);
    }

    @Override
    public int hashCode() {
        return Objects.hash(super.hashCode(), keyIdentifier, personPID, serviceActorPID, modifierPID, personUUID, version, versionDescription, reservedUntil, modifiedTime, status, modifiers, identifiers, names, genders, birthDates, contacts, addresses);
    }
}
