/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package dk.sds.ecpr2.rest;

import dk.sds.ecpr2.dao.ExportHistoryDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.ExportHistoryVO;
import dk.sds.ecpr2.util.TimeHelper;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

/**
 * HistoryExporter
 */
@WebServlet(name="historyExport", urlPatterns={"/historyExport"})
@Service
public class ExportHistoryController extends AbstractServlet {

    @Autowired
    protected ExportHistoryDAO exportHistoryDAO;

    @Value("${jobs.exportHistoryToFile.delayBetween:3600000}")
    private String delayBetweenExportHistory;

    private static final Logger log = LogManager.getLogger(ExportHistoryController.class);

    @Override
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException {
        historyExport(response);
    }

    public void historyExport(HttpServletResponse response) throws IOException{
        String lineBreak = System.getProperty("line.separator");
        String exportNumber = "0";
        String exportTime = "No export made yet";
        String errorMessage = "no error";
        Integer status = 200; // Default is ok
        long timeBetweenExportInMs =  Long.parseLong(delayBetweenExportHistory);

        StringBuilder b = new StringBuilder();

        b.append("Export History:").append(lineBreak).append(lineBreak);
        try {
            ExportHistoryVO exportHistoryVO = exportHistoryDAO.getLatestExport();
            if(exportHistoryVO != null) {
                // If export history has been executed with error or data
                exportNumber = exportHistoryVO.getExportNumber().toString();
                if(exportHistoryVO.getErrorMessage() != null){
                    if (exportHistoryVO.getExportTime().getTime() < TimeHelper.getCurrentTime().getTime()-timeBetweenExportInMs) {
                        // If error is from old export
                        exportHistoryVO = exportHistoryDAO.getLatestSuccesExport();
                    } else {
                        // If error in recent export
                        status = 500;
                        errorMessage = exportHistoryVO.getErrorMessage();
                    }
                }
                if(exportHistoryVO == null){
                    // If error found, and export history only happened with errors.
                    exportNumber = "0";
                } else{
                    DateFormat df = new SimpleDateFormat("MM/dd/yyyy HH:mm:ss");
                    exportTime = df.format(exportHistoryVO.getExportTime());
                }
            }
        } catch (DAOException e) {
            log.error("Export history could not be updated because of an error.", e);
        }

        b.append("Http status: \t\t\t").append(status).append(lineBreak);
        b.append("Error message: \t\t\t").append(errorMessage).append(lineBreak);
        b.append("Export number: \t\t\t").append(exportNumber).append(lineBreak);
        b.append("Tme of last export with data: \t").append(exportTime).append(lineBreak);

        response.getWriter().println(b.toString());
        response.setStatus(status);
    }
}
