/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.security;

import com.sun.xml.messaging.saaj.soap.name.NameImpl;
import dk.sds.ecpr2.util.ECPRException;
import dk.sds.ecpr2.util.FaultCode;
import dk.sds.nsp.security.Security;
import dk.sds.nsp.security.SecurityContext;
import dk.sdsd.dgws._2010._08.PredefinedRequestedRole;
import org.apache.axis.message.Detail;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.annotation.PostConstruct;
import javax.xml.namespace.QName;
import javax.xml.rpc.soap.SOAPFaultException;
import javax.xml.soap.DetailEntry;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;
import java.util.*;

@Component
public class RoleValidationInterceptor implements SOAPHandler<SOAPMessageContext> {

    private static final List<String> NON_INTERCEPTING_SOAP_ACTIONS = Collections.singletonList("http://sds.dk/ecpr2/2017/10/01#GetPermissions");

    private final Map<PredefinedRequestedRole, Authenticator> roleToAuthenticatorMap = new HashMap<>();

    private static final Logger log = LogManager.getLogger(RoleValidationInterceptor.class);

    @Autowired
    AdminAuthenticator adminAuthenticator;
    @Autowired
    DoctorAuthenticator doctorAuthenticator;
    @Autowired
    DentistAuthenticator dentistAuthenticator;
    @Autowired
    MidwifeAuthenticator midwifeAuthenticator;
    @Autowired
    NurseAuthenticator nurseAuthenticator;
    @Autowired
    ECprCoordinatorAuthenticator eCprCoordinatorAuthenticator;
    @Autowired
    SystemAuthenticator systemAuthenticator;
    @Autowired
    AssistantAuthenticator assistantAuthenticator;

    @PostConstruct
    public void postConstruct() {
        roleToAuthenticatorMap.put(PredefinedRequestedRole.WEB_ADMINISTRATOR, adminAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.LÆGE, doctorAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.TANDLÆGE, dentistAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.JORDEMODER, midwifeAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.SYGEPLEJERSKE, nurseAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.E_CPR_KOORDINATOR, eCprCoordinatorAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.SYSTEM, systemAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.ASSISTENT_FOR_JORDEMODER, assistantAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.ASSISTENT_FOR_LÆGE, assistantAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.ASSISTENT_FOR_SYGEPLEJERSKE, assistantAuthenticator);
        roleToAuthenticatorMap.put(PredefinedRequestedRole.ASSISTENT_FOR_TANDLÆGE, assistantAuthenticator);
    }

    public Map<PredefinedRequestedRole, Authenticator> getRoleToAuthenticatorMap() {
        return roleToAuthenticatorMap;
    }

    public OnBehalfOf getOnBehalfOf(SOAPHeader source) {
        OnBehalfOf result = new OnBehalfOf();

        NodeList onBehalfOf = source.getElementsByTagNameNS("urn:sds-dk:ecpr2", "OnBehalfOf");
        if (onBehalfOf.getLength() > 0) {
            Node node = onBehalfOf.item(0);
            NodeList childNodes = node.getChildNodes();
            for (int i = 0; i < childNodes.getLength(); i++) {
                Node item = childNodes.item(i);
                if ("AuthorisationIdentifier".equals(item.getLocalName())) {
                    result.setOnBehalfOfAuthorisationNo(item.getTextContent());
                    break;
                }
                if ("PersonIdentifierType".equals(item.getLocalName())) {
                    NodeList personIdentifierChildren = item.getChildNodes();
                    for (int j = 0; j < personIdentifierChildren.getLength(); j++) {
                        Node personIdentifierChildItem = personIdentifierChildren.item(j);
                        if ("PersonIdentifier".equals(personIdentifierChildItem.getLocalName())) {
                            result.setOnBehalfOfPersonIdentifier(personIdentifierChildItem.getTextContent());
                            break;
                        }
                    }

                    break;
                    // attribute source=CPR is also in the request, we can extend SecurityCredential to use it if needed
                }
            }

        }
        return result;
    }

    protected boolean returnError(String errorMessage) {
        log.error(errorMessage);
        try {
            Detail detail = new Detail();
            DetailEntry detailEntry = detail.addDetailEntry(NameImpl.convertToName(new QName("http://springframework.org/spring-ws", "ValidationError", "spring-ws")));
            detailEntry.addTextNode(errorMessage);
            throw new SOAPFaultException(new QName("http://springframework.org/spring-ws", "ValidationError", "spring-ws"), "Validation error", "", detail);
        } catch (SOAPException e) {
            throw new RuntimeException(e);
        }
    }

    public boolean shouldIntercept(SOAPMessageContext messageContext) {
        String soapAction = ((Map<String, List<String>>) messageContext.get(MessageContext.HTTP_REQUEST_HEADERS)).get("SOAP Action").get(0);
        if (StringUtils.hasLength(soapAction) && soapAction.charAt(0) == '"' && soapAction.charAt(soapAction.length() - 1) == '"') {
            soapAction = soapAction.substring(1, soapAction.length() - 1);
        }
        boolean nonInterceptingAction = NON_INTERCEPTING_SOAP_ACTIONS.contains(soapAction);
        if (nonInterceptingAction) {
            log.debug("Not validating role");
        }
        return !nonInterceptingAction;
    }

    @Override
    public Set<QName> getHeaders() {
        return null;
    }

    @Override
    public boolean handleMessage(SOAPMessageContext soapMessageContext) {
        try {
            log.info("Autentikerer");
            SOAPMessage request = soapMessageContext.getMessage();
            SOAPHeader soapHeader = request.getSOAPHeader();
            NodeList requestedRoleList = soapHeader.getElementsByTagNameNS("http://www.sdsd.dk/dgws/2010/08", "RequestedRole");
            if (requestedRoleList.getLength() == 0) {
                log.warn("Der er ikke medgivet nogen RequestedRole");
                return returnError("Der er ikke medgivet noget RequestedRole");
            }
            String requestedRole = requestedRoleList.item(0).getTextContent();

            if (!shouldIntercept(soapMessageContext)) {
                return true;
            }

            List<String> errorMessages = new ArrayList<>();
            OnBehalfOf onBehalfOf = getOnBehalfOf(soapHeader);
            SecurityContext securityContext = Security.getSecurityContext();
            Optional<SecurityContext.Ticket> ticketOptional = securityContext.getTicket();
            if (!ticketOptional.isPresent()) {
                throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
            }
            SecurityContext.Ticket ticket = ticketOptional.get();
            if (!ticket.isValid()) {
                throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
            }
            if (errorMessages.size() > 0) {
                return returnError(errorMessages.get(0));
            }
            PredefinedRequestedRole predefinedRequestedRole = null;
            try {
                predefinedRequestedRole = PredefinedRequestedRole.fromValue(requestedRole);
            } catch (IllegalArgumentException ignored) {
                // The below if statement handles unknown roles
            }

            Authenticator authenticator = roleToAuthenticatorMap.get(predefinedRequestedRole);
            if (authenticator == null) {
                log.warn("Rollen " + requestedRole + " kan ikke genkendes");
                // Return requestedRole, because that is what the client sent to the backend
                return returnError("Rollen " + requestedRole + " kan ikke genkendes");
            }
            log.info("Autentikerer for rollen " + requestedRole);
            if (!authenticator.authenticate(securityContext, requestedRole, onBehalfOf)) {
                if (securityContext.getActingUser().isPresent() && securityContext.getActingUser().get().getIdentifier() != null) {
                    log.warn(securityContext.getActingUser().get().getIdentifier() + " er ikke en " + requestedRole);
                    // Return requestedRole, because that is what the client sent to the backend
                    return returnError(securityContext.getActingUser().get().getIdentifier() + " er ikke en " + requestedRole);
                } else {
                    log.warn(securityContext.getClient().get().getName() + " er ikke et " + requestedRole);
                    // Return requestedRole, because that is what the client sent to the backend
                    return returnError(securityContext.getClient().get().getName() + " er ikke et " + requestedRole);
                }
            }
            RequestContext.get().setPermissions(authenticator.getPermissions());
            RequestContext.get().setRequestedRole(requestedRole);
            RequestContext.get().setValidatedRole(requestedRole);
            return true;
        } catch (DataAccessException e) {
            log.error("Error: ", e);
            return returnError("Intern server fejl");
        } catch (Exception e) {
            log.error("Error: ", e);
            return returnError(e.getMessage());
        }
    }

    @Override
    public boolean handleFault(SOAPMessageContext soapMessageContext) {
        return true;
    }

    @Override
    public void close(MessageContext messageContext) {

    }
}
