/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.util;

import dk.sds.ecpr2.KeyIdentifier;
import dk.sds.ecpr2.ModificatorPersonType;
import dk.sds.ecpr2.Modified;
import dk.sds.ecpr2.ModifiedBy;
import dk.sds.ecpr2.domain.*;
import dk.sds.nsp.audit.Audit;
import dk.sds.nsp.audit.AuditBuilder;
import dk.sds.nsp.audit.AuditBuilder.TypeOfInformation;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.function.Supplier;

@Component
public class AuditLogger {
    public static final String COMPONENT = "ECPR2";
    public static final String CREATE_PERSON = "createPerson";
    public static final String MERGE_PERSON = "mergePerson";
    public static final String RESERVE_PERSON_IDS = "reservePersonIds";
    private static final String UPDATE_PERSON = "updatePerson";
    private static final String GET_PERSONS_BY_ID = "getPersonsById";
    private static final String DELETE_PERSON = "deletePerson";
    private static final String COMPARE_NAME = "compareName";
    private static final String SEARCH_PERSONS = "searchPersons";
    private static final String GET_PERMISSIONS = "getPermissions";
    private static final String GET_CPR_PERSON = "getCprPerson";

    private final Supplier<AuditBuilder> auditBuilderSupplier;

    public AuditLogger() {
        this(Audit::createAuditBuilder);
    }

    public AuditLogger(Supplier<AuditBuilder> auditBuilderSupplier) {
        this.auditBuilderSupplier = auditBuilderSupplier;
    }

    public void logSearchPersons(String searchQuery, ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        auditLog(auditBuilder, SEARCH_PERSONS, Type.SEARCH_QUERY, searchQuery);

        auditLogActor(auditBuilder, SEARCH_PERSONS, actor);
        auditLogRequestContext(auditBuilder, SEARCH_PERSONS, context);
    }

    public void logCreatePerson(PersonVO person, ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        if (person != null) {
            auditLog(auditBuilder, CREATE_PERSON, Type.PERSON_ID, person.getPersonPID());
            auditLog(auditBuilder, CREATE_PERSON, Type.PERSON_UUID, person.getPersonUUID());
        }
        auditLog(auditBuilder, CREATE_PERSON, Type.PERSON, person);

        auditLogActor(auditBuilder, CREATE_PERSON, actor);
        auditLogRequestContext(auditBuilder, CREATE_PERSON, context);
    }

    public void logMergePerson(KeyIdentifierVO from, KeyIdentifierVO to, Modified modified, ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        auditLog(auditBuilder, MERGE_PERSON, Type.MERGE_PERSON_FROM, from);
        auditLog(auditBuilder, MERGE_PERSON, Type.MERGE_PERSON_TO, to);

        if (modified != null) {
           auditLogModifiedBy(auditBuilder, MERGE_PERSON, modified.getBy());
        }

        auditLogActor(auditBuilder, MERGE_PERSON, actor);
        auditLogRequestContext(auditBuilder, MERGE_PERSON, context);
    }

    public void logReservePersonIds(List<ReserveRequestVO> reserveRequests, ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        for (ReserveRequestVO reserveRequest : reserveRequests) {
            if (reserveRequest == null) continue;
            auditLog(auditBuilder, RESERVE_PERSON_IDS, Type.OID, reserveRequest.getOid());
            auditLog(auditBuilder, RESERVE_PERSON_IDS, Type.RESERVED_UNTIL, reserveRequest.getReservedUntil());
            auditLogModifiedBy(auditBuilder, RESERVE_PERSON_IDS, reserveRequest.getModifiedBy());
        }

        auditLogActor(auditBuilder, RESERVE_PERSON_IDS, actor);
        auditLogRequestContext(auditBuilder, RESERVE_PERSON_IDS, context);
    }

    public void logUpdatePerson(PersonVO person, ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        if (person != null) {
            auditLog(auditBuilder, UPDATE_PERSON, Type.PERSON_ID, person.getPersonPID());
            auditLog(auditBuilder, UPDATE_PERSON, Type.PERSON_UUID, person.getPersonUUID());
        }
        auditLog(auditBuilder, UPDATE_PERSON, Type.PERSON, person);

        auditLogActor(auditBuilder, UPDATE_PERSON, actor);
        auditLogRequestContext(auditBuilder, UPDATE_PERSON, context);
    }

    public void logGetPersonsById(List<PersonVO> persons, ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        for (PersonVO person : persons) {
            if (person == null) continue;
            auditLog(auditBuilder, GET_PERSONS_BY_ID, Type.PERSON_ID, person.getPersonPID());
            auditLog(auditBuilder, GET_PERSONS_BY_ID, Type.PERSON_UUID, person.getPersonUUID());
        }

        auditLogActor(auditBuilder, GET_PERSONS_BY_ID, actor);
        auditLogRequestContext(auditBuilder, GET_PERSONS_BY_ID, context);
    }

    public void logDeletePerson(List<IdentifierVO> identifiers, Modified modified, ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        for (IdentifierVO identifier : identifiers) {
            if (identifier == null) continue;
            auditLog(auditBuilder, DELETE_PERSON, Type.PERSON_ID, identifier.getPersonPID());
            auditLog(auditBuilder, DELETE_PERSON, Type.PERSON_UUID, identifier.getPersonUUID());
        }

        if (modified != null) {
            auditLogModifiedBy(auditBuilder, DELETE_PERSON, modified.getBy());

        }

        auditLogActor(auditBuilder, DELETE_PERSON, actor);
        auditLogRequestContext(auditBuilder, DELETE_PERSON, context);
    }

    public void logCompareName(KeyIdentifier keyIdentifier, String name, ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        if (keyIdentifier != null) {
            auditLog(auditBuilder, COMPARE_NAME, Type.KEY_IDENTIFIER_OID, keyIdentifier.getOID());
            auditLog(auditBuilder, COMPARE_NAME, Type.KEY_IDENTIFIER_VALUE, keyIdentifier.getValue());
        }

        auditLog(auditBuilder, COMPARE_NAME, Type.PERSON_NAME, name);

        auditLogActor(auditBuilder, COMPARE_NAME, actor);
        auditLogRequestContext(auditBuilder, COMPARE_NAME, context);
    }

    public void logGetPermissions(ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        auditLogActor(auditBuilder, GET_PERMISSIONS, actor);
        auditLogRequestContext(auditBuilder, GET_PERMISSIONS, context);
    }

    public void logGetCprPerson(KeyIdentifier keyIdentifier, ServiceActorVO actor, RequestContextVO context) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();

        if (keyIdentifier != null) {
            auditLog(auditBuilder, GET_CPR_PERSON, Type.KEY_IDENTIFIER_OID, keyIdentifier.getOID());
            auditLog(auditBuilder, GET_CPR_PERSON, Type.KEY_IDENTIFIER_VALUE, keyIdentifier.getValue());
        }

        auditLogActor(auditBuilder, GET_PERMISSIONS, actor);
        auditLogRequestContext(auditBuilder, GET_PERMISSIONS, context);
    }

    private void auditLogRequestContext(AuditBuilder auditBuilder, String context, RequestContextVO requestContext) {
        if (requestContext == null) return;

        auditLog(auditBuilder, context, Type.CONTEXT_REQUEST_TIMESTAMP, requestContext.getRequestTimeStamp());
        auditLog(auditBuilder, context, Type.CONTEXT_USER_TYPE, requestContext.getUserType());
        auditLog(auditBuilder, context, Type.CONTEXT_SYSTEM_ID, requestContext.getSystemIdentifier());
    }

    private void auditLogActor(AuditBuilder auditBuilder, String context, ServiceActorVO actor) {
        if (actor == null) return;

        auditLog(auditBuilder, context, Type.ACTOR_NAME, actor.getName());
        auditLog(auditBuilder, context, Type.ACTOR_ID, actor.getPersonIdentifier());
        auditLog(auditBuilder, context, Type.ACTOR_ROLE, actor.getRole());
        auditLog(auditBuilder, context, Type.ACTOR_ORGANIZATION_ID, actor.getOrganizationId());
        auditLog(auditBuilder, context, Type.ACTOR_ORGANIZATION_NAME, actor.getOrganizationName());
        auditLog(auditBuilder, context, Type.ACTOR_AUTHORISATION_NUMBER, actor.getAuthorisationNumber());
    }

    private void auditLogModifiedBy(AuditBuilder auditBuilder, String context, ModifiedBy by) {
        if (by == null) return;

        if(by.getPerson() != null) {
            ModificatorPersonType modifiedBy = by.getPerson();
            auditLog(auditBuilder, context, Type.MODIFIED_BY_NAME, modifiedBy.getName());
            auditLog(auditBuilder, context, Type.MODIFIED_BY_ID, modifiedBy.getPersonIdentifier());
            auditLog(auditBuilder, context, Type.MODIFIED_BY_AUTHORISATION_ID, modifiedBy.getAuthorisationIdentifier());
        }

        if(by.getAuthorisedBy() != null && by.getAuthorisedBy().getPerson() != null) {
            ModificatorPersonType authorisedBy = by.getAuthorisedBy().getPerson();
            auditLog(auditBuilder, context, Type.AUTHORISED_BY_NAME, authorisedBy.getName());
            auditLog(auditBuilder, context, Type.AUTHORISED_BY_ID, authorisedBy.getPersonIdentifier());
            auditLog(auditBuilder, context, Type.AUTHORISED_BY_AUTHORISATION_ID, authorisedBy.getAuthorisationIdentifier());
        }
    }

    private void auditLog(AuditBuilder auditBuilder, String context, Type type, Object value) {
        if (value == null) {
            value = "null";
        }
        auditBuilder.addAuditInformation(COMPONENT, context, type.typeOfInformation, type.value, value);
    }


    enum Type {
        PERSON("person", TypeOfInformation.RegularPersonalInformation),
        PERSON_ID("person-id", TypeOfInformation.RegularPersonalInformation),
        PERSON_UUID("person-id-uuid", TypeOfInformation.RegularPersonalInformation),
        PERSON_NAME("person-name", TypeOfInformation.RegularPersonalInformation),
        MERGE_PERSON_FROM("merge-person-from", TypeOfInformation.RegularPersonalInformation),
        MERGE_PERSON_TO("merge-person-do", TypeOfInformation.RegularPersonalInformation),
        ACTOR_ID("actor-id", TypeOfInformation.NonPersonalInformation),
        ACTOR_ROLE("actor-role", TypeOfInformation.NonPersonalInformation),
        ACTOR_ORGANIZATION_NAME("actor-organization-name", TypeOfInformation.NonPersonalInformation),
        ACTOR_ORGANIZATION_ID("actor-organization-id", TypeOfInformation.RegularPersonalInformation),
        ACTOR_AUTHORISATION_NUMBER("actor-authorisation-number", TypeOfInformation.NonPersonalInformation),
        CONTEXT_USER_TYPE("context-user-type", TypeOfInformation.NonPersonalInformation),
        CONTEXT_SYSTEM_ID("context-system-id", TypeOfInformation.NonPersonalInformation),
        CONTEXT_REQUEST_TIMESTAMP("context-request-timestamp", TypeOfInformation.NonPersonalInformation),
        KEY_IDENTIFIER_OID("key-identifier-oid", TypeOfInformation.NonPersonalInformation),
        KEY_IDENTIFIER_VALUE("key-identifier", TypeOfInformation.RegularPersonalInformation),
        OID("oid", TypeOfInformation.NonPersonalInformation),
        RESERVED_UNTIL("reserved-until", TypeOfInformation.NonPersonalInformation),
        ACTOR_NAME("actor-name", TypeOfInformation.RegularPersonalInformation),
        MODIFIED_BY_NAME("modifier-name", TypeOfInformation.RegularPersonalInformation),
        MODIFIED_BY_ID("modifier-id", TypeOfInformation.RegularPersonalInformation),
        MODIFIED_BY_AUTHORISATION_ID("modifier-authorisation_id", TypeOfInformation.NonPersonalInformation),
        AUTHORISED_BY_NAME("authoriser-name", TypeOfInformation.RegularPersonalInformation),
        AUTHORISED_BY_ID("authoriser-id", TypeOfInformation.RegularPersonalInformation),
        AUTHORISED_BY_AUTHORISATION_ID("authoriser-authorisation-id", TypeOfInformation.NonPersonalInformation),
        SEARCH_QUERY("search-query", TypeOfInformation.NonPersonalInformation);

        private final String value;
        private final TypeOfInformation typeOfInformation;

        Type(String value, TypeOfInformation typeOfInformation) {
            this.value = value;
            this.typeOfInformation = typeOfInformation;
        }
    }
}
