/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.ObjectFactory;
import dk.sds.ecpr2._2017._10._01.*;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.security.Permission;
import dk.sds.ecpr2.security.RequestContext;
import dk.sds.ecpr2.service.ECPR2ServiceImpl;
import dk.sds.ecpr2.util.*;
import dk.sds.ecpr2.util.Properties;
import dk.sds.ecpr2.webservice.mapping.FromXMLMapper;
import dk.sds.ecpr2.webservice.mapping.ToXMLMapper;
import dk.sds.ecpr2.webservice.validation.RequestValidator;
import dk.sds.nsp.security.Security;
import dk.sds.nsp.security.SecurityContext;
import org.apache.cxf.annotations.SchemaValidation;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.jws.WebService;
import javax.servlet.annotation.WebServlet;
import javax.xml.ws.BindingType;
import java.math.BigInteger;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

import static javax.xml.ws.soap.SOAPBinding.SOAP11HTTP_BINDING;

@WebService(name = "eCPR2Service", targetNamespace = "http://sds.dk/ecpr2/2017/10/01/")
@WebServlet(name = "eCPR2Service", urlPatterns = {"/services/ecpr2/"})
@BindingType(value = SOAP11HTTP_BINDING)
@SchemaValidation(type = SchemaValidation.SchemaValidationType.BOTH)
@Service
public class ECPR2WebServiceImpl extends AbstractWebService implements ECPR2PortType {

    private static final Logger log = LogManager.getLogger(ECPR2WebServiceImpl.class);

    protected class RequestMetaData {
        private ServiceActorVO actor;
        private RequestContextVO context;

        public ServiceActorVO getActor() {
            return actor;
        }

        public void setActor(ServiceActorVO actor) {
            this.actor = actor;
        }

        public RequestContextVO getContext() {
            return context;
        }

        public void setContext(RequestContextVO context) {
            this.context = context;
        }
    }

    @Autowired
    private ECPR2ServiceImpl businessService;

    ECPR2ServiceImpl getBusinessService() {
        return businessService;
    }

    @Autowired
    FromXMLMapper requestMapper;

    @Autowired
    ToXMLMapper responseMapper;

    @Value("${admin.service.shouldCheckPermissions:true}")
    private boolean shouldCheckPermissions;

    @Autowired
    Properties properties;

    @Autowired
    AuditLogger auditLogger;

    @Override
    public SearchPersonResponse searchPerson20171001(SearchPersonRequest request) throws DGWSFault {
        setupSlalogItem();
        checkPermissions(Permission.Read);
        final SearchPersonResponse response = new SearchPersonResponse();
        log.info("Request: " + request.getQuery());
        try {
            new RequestValidator().validate(request);

            String searchQuery = request.getQuery();
            int totalPersonAmount = businessService.countPersons(searchQuery);
            if (request.getMaxNumberOfResults() == null || totalPersonAmount <= request.getMaxNumberOfResults().intValue()) {
                List<PersonVO> personList = businessService.searchPersons(searchQuery);
                response.getPersons().addAll(mapPersons(personList));
            }
            response.setNumberOfPersons(BigInteger.valueOf(totalPersonAmount));
            log.info("Results: " + response.getPersons().size());
            RequestMetaData metaData = getRequestMetadata();
            auditLogger.logSearchPersons(searchQuery, metaData.getActor(), metaData.getContext());

        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.SEARCH_PERSON_VALIDATION_ERROR, FaultCode.SEARCH_PERSON_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    public GetPersonByIdResponse getPersonById20171001(GetPersonByIdRequest request) throws DGWSFault {
        setupSlalogItem();
        checkPermissions(Permission.Read);
        final GetPersonByIdResponse response = new GetPersonByIdResponse();
        try {
            new RequestValidator().validate(request);
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            List<PersonVO> personList = requestMapper.mapIdentifierToPersonFromRequest(request.getKeyIdentifiers());
            List<PersonVO> returnedPersonList = businessService.getPersonsById(personList, metaData.getActor(), metaData.getContext());
            response.getPersons().addAll(mapPersons(returnedPersonList));
            auditLogger.logGetPersonsById(personList, metaData.getActor(), metaData.getContext());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_PERSON_BY_PERSONID_VALIDATION_ERROR, FaultCode.GET_ECPR_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    public CreatePersonResponse createPerson20171001(CreatePersonRequest request) throws DGWSFault {
        setupSlalogItem();
        checkPermissions(Permission.Write);
        final CreatePersonResponse response = new CreatePersonResponse();
        try {
            new RequestValidator().validate(request);
            PersonVO personVO = requestMapper.mapPersonFromRequest(request.getPerson());
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            PersonVO createdPersonVO = businessService.createPerson(personVO, metaData.getActor(), metaData.getContext());
            PersonType createdPerson = responseMapper.mapPerson(createdPersonVO);
            response.setPerson(createdPerson);
            auditLogger.logCreatePerson(personVO, metaData.getActor(), metaData.getContext());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.CREATE_ECPR_VALIDATION_ERROR, FaultCode.CREATE_ECPR_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    public ReservePersonIdResponse reservePersonId20171001(ReservePersonIdRequest request) throws DGWSFault {
        setupSlalogItem();
        checkPermissions(Permission.Write);
        final ReservePersonIdResponse response = new ReservePersonIdResponse();

        try {
            new RequestValidator().validate(request);
            List<ReserveRequestVO> reserveRequests = requestMapper.mapReserveRequest(request.getLists());
            reserveRequests.forEach(r -> r.setModifiedBy(request.getModified().getBy()));
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            List<List<PersonVO>> personsList = businessService.reservePersonIds(reserveRequests, metaData.getActor(), metaData.getContext());
            responseMapper.mapReserveResponse(response, personsList);
            auditLogger.logReservePersonIds(reserveRequests, metaData.getActor(), metaData.getContext());

        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.CREATE_ECPR_VALIDATION_ERROR, FaultCode.CREATE_ECPR_INTERNAL_ERROR);
        }

        return response;
    }

    @Override
    public UpdatePersonResponse updatePerson20171001(UpdatePersonRequest request) throws DGWSFault {
        setupSlalogItem();
        checkPermissions(Permission.Write);
        final UpdatePersonResponse response = new UpdatePersonResponse();
        try {
            new RequestValidator().validate(request);
            PersonVO personUpdate = requestMapper.mapPersonFromRequest(request.getPerson());
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            PersonVO person = businessService.updatePerson(personUpdate, metaData.getActor(), metaData.getContext());
            PersonType updatedPerson = responseMapper.mapPerson(person);
            response.setPerson(updatedPerson);
            auditLogger.logUpdatePerson(personUpdate, metaData.getActor(), metaData.getContext());

        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.UPDATE_ECPR_VALIDATION_ERROR, FaultCode.UPDATE_ECPR_INTERNAL_ERROR);
        }

        return response;
    }

    @Override
    public MergePersonsResponse mergePersons20171001(MergePersonsRequest request) throws DGWSFault {
        setupSlalogItem();
        checkPermissions(Permission.Write);
        final MergePersonsResponse response = new MergePersonsResponse();
        try {
            new RequestValidator().validate(request);
            KeyIdentifierVO from = requestMapper.mapKeyIdentifier(request.getFrom());
            KeyIdentifierVO to = requestMapper.mapKeyIdentifier(request.getTo());
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            Modified modified = request.getModified();
            PersonVO person = businessService.mergePerson(from, to, modified, metaData.getActor(), metaData.getContext());
            PersonType mergedPerson = responseMapper.mapPerson(person);
            response.setPerson(mergedPerson);
            auditLogger.logMergePerson(from, to, modified, metaData.getActor(), metaData.getContext());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.MERGE_PERSON_VALIDATION_ERROR, FaultCode.MERGE_PERSON_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    public DeletePersonResponse deletePerson20171001(DeletePersonRequest request) throws DGWSFault {
        setupSlalogItem();
        if (properties.isProduction) {
            FaultUtil.throwECPR2Fault(FaultCode.DELETE_PERSON_BY_PERSONID_NOT_ALLOWED_IN_PROD);
        }
        checkPermissions(Permission.Write);
        final DeletePersonResponse response = new DeletePersonResponse();
        try {
            new RequestValidator().validate(request);
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            List<IdentifierVO> identifiers = requestMapper.mapPersonKeyIdentifier(request.getPersons());
            Modified modified = request.getModified();
            businessService.deletePerson(identifiers, modified, metaData.getActor(), metaData.getContext());
            auditLogger.logDeletePerson(identifiers, modified, metaData.getActor(), metaData.getContext());

        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.DELETE_PERSON_BY_PERSONID_VALIDATION_ERROR, FaultCode.DELETE_PERSON_BY_PERSONID_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    public GetPermissionsResponse getPermissions20171001(GetPermissionsRequest request) throws DGWSFault {
        setupSlalogItem();
        final GetPermissionsResponse response = new GetPermissionsResponse();
        try {
            List<RolesPermissions> responseRolesPermissions = response.getRolesPermissions();
            Map<String, List<Permission>> permissionMap = businessService.getPermissions(request.getGetAllPermissions() != null, request.getGetCallersPermissions() != null);
            permissionMap.forEach((role, permissions) -> {
                RolesPermissions rolesPermissions = new RolesPermissions();
                rolesPermissions.setRequestedRole(role);
                List<String> permissionsList = rolesPermissions.getPermissions();
                permissionsList.addAll(permissions.stream().map(Permission::getName).collect(Collectors.toList()));
                responseRolesPermissions.add(rolesPermissions);
            });
            RequestMetaData metaData = getRequestMetadata();
            auditLogger.logGetPermissions(metaData.getActor(), metaData.getContext());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_PERMISSIONS_VALIDATION_ERROR, FaultCode.GET_PERMISSIONS_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    public GetCprPersonResponse getCprPerson20171001(GetCprPersonRequest request) throws DGWSFault {
        setupSlalogItem();
        checkPermissions(Permission.Read);
        final GetCprPersonResponse response = new GetCprPersonResponse();
        try {
            CprPersonVO cprPersonVO = businessService.getCprPerson(request.getKeyIdentifier());
            if (cprPersonVO != null) {
                CprPerson cprPerson = new CprPerson();
                String name = (cprPersonVO.getGivenName() == null ? "" : cprPersonVO.getGivenName() + " ") + (cprPersonVO.getMiddleName() == null ? "" : cprPersonVO.getMiddleName() + " ") + (cprPersonVO.getSurName() == null ? "" : cprPersonVO.getSurName());
                cprPerson.setName(name.trim());
                cprPerson.setBirthDate(cprPersonVO.getBirthdate());
                response.setCprPerson(cprPerson);
            }

            RequestMetaData metaData = getRequestMetadata();
            auditLogger.logGetCprPerson(request.getKeyIdentifier(), metaData.getActor(), metaData.getContext());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_CPR_PERSON_VALIDATION_ERROR, FaultCode.GET_CPR_PERSON_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    public CompareNameResponse compareName20171001(CompareNameRequest request) throws DGWSFault {
        setupSlalogItem();
        checkPermissions(Permission.Read);
        final CompareNameResponse response = new CompareNameResponse();
        try {
            RequestMetaData metaData = getRequestMetadata();
            Integer score = businessService.compareName(request.getKeyIdentifier(), request.getName(), metaData.getActor(), metaData.getContext());
            if (score != null) {
                response.setScore(score);
            }
            auditLogger.logCompareName(request.getKeyIdentifier(), request.getName(), metaData.getActor(), metaData.getContext());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.COMPARE_NAME_VALIDATION_ERROR, FaultCode.COMPARE_NAME_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    public GetOIDsResponse getOIDs20171001(GetOIDsRequest request) throws DGWSFault {
        setupSlalogItem();
        final GetOIDsResponse response = new ObjectFactory().createGetOIDsResponse();
        try {
            if (request != null) {
                List<OIDVO> list = businessService.getOIDs();
                response.getOIDS().addAll(responseMapper.mapOIDs(list));
            }
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GENERAL_ERROR, FaultCode.GENERAL_ERROR);
        }
        return response;
    }

    private void checkPermissions(Permission... permissions) {
        if (shouldCheckPermissions) {
            boolean permissionsFound = true;

            try {
                for (Permission permission : permissions) {
                    if (!RequestContext.get().getPermissions().contains(permission)) {
                        permissionsFound = false;
                        break;
                    }
                }
            } catch (Exception ex) {
                log.error("Error checking permissions", ex);
                FaultUtil.throwECPR2Fault(FaultCode.AUTHORIZATION_INTERNAL_ERROR, ex);
            }

            if (!permissionsFound) {
                FaultUtil.throwECPR2Fault(FaultCode.AUTHORIZATION_ERROR);
            }
        }
    }

    private void validateSecurityContext() {
        SecurityContext securityContext = Security.getSecurityContext();
        Optional<SecurityContext.Ticket> ticketOptional = securityContext.getTicket();
        if (!ticketOptional.isPresent()) {
            throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
        }
        SecurityContext.Ticket ticket = ticketOptional.get();
        if (!ticket.isValid()) {
            throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
        }
        checkDgws(securityContext);
    }

    private void checkDgws(SecurityContext securityContext) {
        Optional<SecurityContext.User> userOptional = securityContext.getActingUser();
        if (!userOptional.isPresent()) {
            return;
        }
        if (userOptional.get().getUserType().equals(SecurityContext.User.UserType.Citizen)) {
            throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
        }
        final Optional<SecurityContext.User.Credentials> credentials = userOptional.get().getCredentials();
        if (!credentials.isPresent()) {
            throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
        }
    }

    private RequestMetaData getRequestMetadata() {
       final RequestMetaData metaData = new RequestMetaData();
       SecurityContext securityContext = Security.getSecurityContext();
       metaData.setActor(getServiceActor(securityContext));
       metaData.setContext(getRequestContext(securityContext));
       return metaData;
    }

    private RequestContextVO getRequestContext(SecurityContext securityContext) {
        RequestContextVO context = new RequestContextVO();
        if (securityContext.getMessage().isPresent()) {
            SecurityContext.Message message = securityContext.getMessage().get();
            if (message.getConversationIdentifier().isPresent()) {
                context.setFlowId(message.getConversationIdentifier().get());
            }
            if (message.getIdentifier().isPresent()) {
                context.setMessageId(message.getIdentifier().get());
            }
        }
        if (securityContext.getClient().isPresent()) {
            SecurityContext.Client client = securityContext.getClient().get();
            if (client.getName().isPresent()) {
                context.setSystemIdentifier(client.getName().get());
            }
        }
        context.setRequestTimeStamp(Instant.now());
        return context;
    }

    private ServiceActorVO getServiceActor(SecurityContext context) {
        ServiceActorVO actor = new ServiceActorVO();

        Optional<SecurityContext.Organisation> organisation = context.getOrganisation();
        if (organisation.isPresent()) {
            actor.setOrganizationId(organisation.get().getIdentifier());
            if (organisation.get().getName().isPresent()) {
                actor.setOrganizationName(organisation.get().getName().get());
            }
        }
        Optional<SecurityContext.User> user = context.getActingUser();
        if (user.isPresent()) {
            actor.setPersonIdentifier(user.get().getIdentifier());
            if (user.get().getGivenName().isPresent()) {
                actor.setFirstName(user.get().getGivenName().get());
            }
            if (user.get().getSurname().isPresent()) {
                actor.setLastName(user.get().getSurname().get());
            }
            Optional<SecurityContext.User.Credentials> credentials = user.get().getCredentials();
            if (credentials.isPresent()) {
                if (credentials.get().getEducationCode().isPresent()) {
                    actor.setRole(credentials.get().getEducationCode().get()); // Uddannelseskode
                }
                if (credentials.get().getAuthorizationCode().isPresent()) {
                    actor.setAuthorisationNumber(credentials.get().getAuthorizationCode().get().trim());
                }
            }
        }

        return actor;
    }

    private Collection<? extends PersonType> mapPersons(List<PersonVO> personList) throws DAOException {
        List<PersonType> searchPerson = new ArrayList<>();
        for (PersonVO person : personList) {
            searchPerson.add(responseMapper.mapPerson(person));
        }
        return searchPerson;
    }
}
